/*
** Module   : ScrollableSelectionListGuiImpl.java
** Abstract : scrollable selection list implementation. 
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150423 Initial version.
** 002 EVL 20150514 Changing scrollbar painting approach.  Several minor UI fixes.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.Color;
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.Position;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Scrollable selection list GUI implementation.  To be used to support combo box and selection
 * list widgets. 
 */
@SuppressWarnings("unchecked")
public class ScrollableSelectionListGuiImpl
extends ScrollableList<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Widget configuration */
   private BaseConfig config;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Row height */
   private int rowHeight;

   /** The vertical scroll bar associated with this scrollable list. */
   private ScrollBar<GuiOutputManager> vbar;

   /** Selected item box color. */
   private Color selectionBack;

   /** Selected item text color. */
   private Color selectionFore;

   /** Selected dotted line color. */
   private Color selectionDots;
   
   /** The curent top row in the list to draw. */
   private int rowTop;
   
   /**
    * Create component using given configuration and data model.
    * 
    * @param    box
    *           The instance with which this object must be associated.
    * @param    model
    *           The data model that represents the contents of the list.
    */
   ScrollableSelectionListGuiImpl(ComboBox<GuiOutputManager> box, ListModel<String> model)
   {
      super(box, model);
      // initialize screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // and font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, box.config(), false);
      // color initializing
      selectionBack = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      selectionFore = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      // top row number will be initialized later
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      super.requestFocus();
   }

   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public void draw()
   {
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = box.config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_MENU, SysColor.COLOR_BTNTEXT);

      // determine colors to use
      gc.colors();

      // get actual rows to view
      final int rowsVisible = visibleRows();
      // set the position and range for the scrollbar
      if (rowsVisible < model().size())
      {
         vbar._setVisible(true);
         vbar.setMax(model().size());
         vbar.setVisibleRange(rowsVisible);
         vbar.setPosition(currentRow());
      }
      else
      {
         vbar._setVisible(false);
      }
      
      // resolve font
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final int rowHeight = screen().coordinates().heightToNative(1.0); 
      d.height = rowsVisible * rowHeight;

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);

      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            // draw entry field text
            gd.setGuiFont(fntCache.font);
            gd.setFontStyle(fntCache.font.getStyle());
            
            int x = po.x;
            int y = po.y;
            int fontSize = fntCache.font.getSize();
            // adjust top row number to have last item from the list as bottom number 
            if ((rowTop + rowsVisible) > model().size())
            {
               rowTop = model().size() - rowsVisible; 
            }
            
            for (int i = rowTop; i < model().size() && i < rowsVisible + rowTop;
                            i++, y+= rowHeight)
            {
               if (y < 0)
                  continue;
               
               // draw nex visible item in the list          
               drawItem(i, x, y, d.width, rowHeight, fontSize);
            }
         }
         
      });
   }

   /**
    * Process scroll event.
    * 
    * @param   event
    *          Scroll event.
    */
   @Override
   public void processScrollEvent(ScrollEvent event)
   {
      // get the new row to be set as current in the list
      super.ensureRowVisible((int) event.limit().y, false);
      
      // recalc top row changes
      if (currentRow() < rowTop)
      {
         // upper than visible region
         rowTop = currentRow(); 
      }
      else if (currentRow() >= rowTop + visibleRows())
      {
         // lower than visible region
         rowTop = currentRow() - visibleRows() + 1;
      }
      
      // and repaint the list
      ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
      {
         @Override
         public void run()
         {
            repaint();
         }
      });
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      int rowHeight = screen().coordinates().heightToNative(1.0);
      
      synchronized (this)
      {
         if (model().size() > 0)
         {
            NativePoint n = translate(this, e);
            
            // calculate clicked line
            int row = n.y / rowHeight + rowTop;
            
            if (!selected(row))
            {
               // go to selected item
               super.ensureRowVisible(row, false);       
               // and simulate ENTER key to select it
               KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Key.VK_ENTER);
               EventManager.postEvent(ki);
            }
            
            repaint(); 
         }
      }

      ThinClient.getInstance().getFocusManager().setFocusOn(this);
   }

   /**
    * Set the list's vertical scrollarr.
    * 
    * @param    vbar
    *           The vertical scrollbar.
    */
   public void setScrollBar(ScrollBar<GuiOutputManager> vbar)
   {
      this.vbar = vbar;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return super.width() + 1.75;
   }

   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      
      // nothing to do in GUI mode 
   }
   
   /**
    * Set item number to be displayed on the top of the drop-down list.
    * 
    * @param   index
    *          The index of the top item.
    */
   @Override
   protected void setTopRow(int index)
   {
      ensureRowVisible(index, false);
      rowTop = index;
   }
   
   /**
    * Draws the next text item in the list inside visible part of the drop-down.
    * 
    * @param    number
    *           The number of the item to draw.
    * @param    x
    *           The X coordinate of the enclosing rectangle to draw item.
    * @param    y
    *           The Y coordinate of the enclosing rectangle to draw item.
    * @param    rw
    *           The pixel width of the enclosing rectangle to draw item.
    * @param    rh
    *           The pixel height of the enclosing rectangle to draw item.
    * @param    fontSize
    *           The vertical size of font to draw item text.
    */
   private void drawItem(int number, int x, int y, int rw, int rh, int fontSize)
   {
      if (number == currentRow())
      {
         gd.setColor(selectionBack);
         gd.fillRect(x, y, rw, rh);
         gd.setColor(selectionFore);
      }
      else
      {
         gd.setColor(gc.fgColor);
      }

      gd.drawString(getItemText(number), x + ComboBoxGuiImpl.TEXT_SHIFT, y + (rh + fontSize)/2);
   }
}
