/*
** Module   : WinKeyboardReader.java
** Abstract : process swing key events and convert them to 4GL keys, compatible with Windows 
**            clients.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141029 Created initial version.
** 002 CA  20141125 Added CTRL-BREAK processing for GUI client.  On Windows, CTRL-BREAK produces
**                  the CTRL-C key (code 3); on linux, it produces the CTRL-CTRL-S key - CTRL 
**                  modifier with the CTRL-S key (code 19).
** 003 CA  20141205 Added support for the SHIFT modifier.
** 004 CA  20141207 Improved H003 - the SHIFT modifier must be treated only if the modified key 
**                  (SHIFT-<key>) ends with the same label as the unmodified key.  This is needed
**                  to distinguish keys which have a double-role (as uppercase letters or SHIFT 
**                  with a physical key which can emit two characters, depending on if SHIFT was
**                  pressed or not, like the ";:" key in a US layout keyboard).
** 005 CA  20141220 Ignore the SHIFT processing for the SHIFT-TAB key (this is treated explicitly
**                  by keyPressed).
** 006 EVL 20150518 Removing generation of the ESC key event in keyTyped() method.  The ESC key
**                  is a END-ERROR generation and should be invoked only once per key press.  In
**                  GUI both keyPressed() and keyTyped() executed and this will generate two
**                  END-ERROR events and in case of the drop-down dismissing causes the
**                  application to stop after dismissing drop-down.
*/

package com.goldencode.p2j.ui.client.swing;

import java.awt.event.KeyEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * This class represents a {@link SwingKeyboardReader} implementation for clients which require
 * Windows compatibility.
 */
public class WinKeyboardReader
extends SwingKeyboardReader
{
   /**
    * Default c'tor.  It translates specific GUI related keys.
    */
   public WinKeyboardReader()
   {
      xlate.put(KeyEvent.VK_ESCAPE, Key.VK_ESCAPE);
      // xlate.put(KeyEvent.VK_ALT, 1024); // ALT is not processed by 4GL...
   }

   /**
    * This method is called when any key is pressed or when a multi-key combination (CTRL, SHIFT, 
    * ALT modifiers) is pressed.  Key events provided here are low level events and have not been
    * processed into a higher level code as is done in {@link #keyTyped}, with an exception: if
    * CTRL is pressed with a latin character, then the processing is deferred to {@link #keyTyped};
    * otherwise, if CTRL is pressed, bit 11 (CTRL modifier) is set and the key is sent to the
    * {@link SwingKeyboardReader#queue}, in FIFO order.
    * <p>
    * All other ALT processing is deferred to the {@link #keyTyped} method, with this exception:
    * if the 4GL key code is VK_TAB and the SHIFT modifier key is on (the shift key was pressed 
    * at the time the tab was pressed), then a VK_BACKTAB is posted to the queue.  Otherwise, 
    * the VK_TAB is ignored because it is naturally processed in {@link #keyTyped} (but the back 
    * tab is not).
    *  
    * @param    evt
    *           The high level key event to process.
    */
   public synchronized void keyPressed(KeyEvent evt)
   {
      int     code  = -1;
      int     key   = evt.getKeyCode();
      boolean shift = evt.isShiftDown();
      boolean alt   = evt.isAltDown();
      boolean ctrl  = evt.isControlDown();

      Integer mappedCode = xlate.get(key);

      if (ctrl && !ctrlTyped(key))
      {
         if (key != KeyEvent.VK_CONTROL)
         {
            if (key == Key.VK_CTRL_C || key == Key.VK_CTRL_S)
            {
               // on windows, physical CTRL-BREAK emits code for the CTRL-C key (code 3);
               // on linux, physical CTRL-BREAK emits code for CTRL-CTRL-S (CTRL modifier with the 
               // CTRL-S key - code 19)
               
               // either of these need to be translated into 4GL's CTRL-BREAK (code 541)t 
               code = Key.VK_CTRL_BREAK;
            }
            else
            {
               code = mappedCode == null ? key : mappedCode;

               code = code | 1 << 11;
            }
            
            queue.add(code);
         }
         
         evt.consume();
         return;
      }
      
      if (mappedCode != null)
      {
         code = mappedCode.intValue();
         
         // normal VK_TAB is generated naturally in keyTyped() but SHIFT +
         // TAB must be handled here
         if (code == Key.VK_TAB)
         {
            code = shift ? Key.VK_BACK_TAB : -1;
         }
         else if (code == Key.VK_ENTER)
         {
            if (alt)
            {
               code = processAlt(code);
            }
         }
         else
         {
            if (shift)
            {
               code = processShift(code);
            }
         }
         
         if (code != -1)
         {
            queue.add(code);
         }
      }
      
      /*
      if (code == -1)
      {
         System.out.printf("%s\n", evt.toString());
      }
      */
   }
   
   /**
    * This method is called when a pressed key (or a multi-key combination) is converted to a 
    * higher level representation. As a general rule, all key codes between 0 and 127 inclusive 
    * are generated through this callback. The reason for this is that Java translates the keys or
    * multi-key combinations into the higher level representation which almost perfectly matches 
    * the 4GL key codes 0 through 127.  In other words, the keyboard inputs of the user are 
    * very compatible between the 4GL and Java for those portions of the key codes (0 through 127).
    * <p>
    * Once a valid key code is generated, that key code is added to the key input queue in FIFO 
    * order.
    * <p>
    * The non-visual values 0 through 31 are handled almost "for free". The user generates these 
    * in the following manner:
    * <p>
    * <pre>
    * Key Code  User Input               Notes
    * --------  -----------------------  --------------------------------------
    * 0         CTRL+2                   Special code in this method supports this.
    * 1         CTRL+A
    * 2         CTRL+B
    * 3         CTRL+C
    * 4         CTRL+D
    * 5         CTRL+E
    * 6         CTRL+F
    * 7         CTRL+G
    * 8         CTRL+H or VK_BACKSPACE
    * 9         CTRL+I or VK_TAB         The containing frame's focus traversal
    *                                    keys must be forced to the empty set in
    *                                    order for this to work.
    * 10        CTRL+J or VK_ENTER       NOT DELIVERED, handled in keyPressed()
    * 11        CTRL+K
    * 12        CTRL+L
    * 13        CTRL+M
    * 14        CTRL+N
    * 15        CTRL+O
    * 16        CTRL+P
    * 17        CTRL+Q
    * 18        CTRL+R
    * 19        CTRL+S
    * 20        CTRL+T
    * 21        CTRL+U
    * 22        CTRL+V
    * 23        CTRL+W
    * 24        CTRL+X
    * 25        CTRL+Y
    * 26        CTRL+Z
    * 27        VK_ESCAPE                DELIVERED; Windows mode allows this.
    * 28        CTRL+\
    * 29        CTRL+]
    * 30        CTRL+6                   Special code in this method supports this.
    * 31        CTRL+-                   Special code in this method supports this.
    * </pre>
    * <p>
    * ALT key is the modifier which sets the key code's 10th bit.  If this is determined to be
    * pressed together with the key, than bit 10 will be set and this key will be pushed to the
    * {@link SwingKeyboardReader#queue}, but only if the key code's label is not the empty string.
    * 
    * @param    evt
    *           The high level key event to process.
    */
   public synchronized void keyTyped(KeyEvent evt)
   {
      if (evt.isConsumed())
      {
         return;
      }

      boolean ctrl = evt.isControlDown();
      boolean alt  = evt.isAltDown();
      boolean shift = evt.isShiftDown();
      char    ch   = evt.getKeyChar();
      int     code = -1;
      
      // all CTRL combinations for codes 0 through 31 inclusive are normally
      // provided by the TYPED event, except the following:
      if (ctrl)
      {
         if (ch == '2')
         {
            // CTRL-@
            code = 0;
         }
         else if (ch == '6')
         {
            // CTRL-^
            code = 30;
         }
         else if (ch == '-')
         {
            // CTRL-_
            code = 31;
         }
      }
      
      // normal processing: the preprocessed character is returned directly
      if (code == -1)
      {
         // ALT/ENTER are processed in keyPressed() and not here
         if (ch != KeyEvent.VK_ALT && ch != KeyEvent.VK_ENTER)
         {
            code = (int) ch;
            
            if (alt)
            {
               // Remapping lowercase Latin to uppercase. This differs
               // with Progress but seems to be more convenient to user because
               // Ctrl+<letter> combinations are case insensitive (and even
               // keyboard layout insensitive). Note the usage of direct
               // calculation not Character class methods. This is
               // because we are dealing with integer key codes not characters
               // so this is dirty hack anyway.
               if (code >= 'a' && code <= 'z')
               {
                  code -= 'a' - 'A';
               }
            }
         }
      }
      
      if (code != -1)
      {
         // honor the alt modifier if it was on previously
         if (alt)
         {
            code = processAlt(code);
         }
         
         if (shift)
         {
            code = processShift(code);
         }
         
         // if alt processing disabled the key, exit now
         
         // also for GUI terminal ESC key should be processed only once in keyPressed()
         // otherwise the application will stop when the drop-down list dismissing by ESC
         if (code == -1 || code == Key.VK_ESCAPE)
            return;
         
         queue.add(new Integer(code));
      }
   }

   /**
    * Process the given key by setting the 10th bit.
    *
    * @param    code
    *           The key code to be processed.
    *           
    * @return   The processed key code or <code>-1</code> if the ALT-key combination has no label.
    */
   private int processAlt(int code)
   {
      // modify the next character (in GUI 4GL, ALT is a modifier key)
      code = code | 1 << 10;
      
      if (Keyboard.keyLabel(code).length() == 0)
      {
         code = -1;
         queue.add(code);
      }
      
      return code;
   }
   
   /**
    * Process the given key by setting the 9th bit.
    *
    * @param    code
    *           The key code to be processed.
    *           
    * @return   The processed key code or <code>-1</code> if the SHIFT-key combination has no label.
    */
   private int processShift(int code)
   {
      // ignore tab - this is treated separately
      if (code == KeyEvent.VK_TAB)
      {
         return -1;
      }

      int orig = code;
      String lblOrig = Keyboard.keyLabel(orig);
      
      // modify the next character (in GUI 4GL, SHIFT is a modifier key)
      code = code | 1 << 9;
      String lbl = Keyboard.keyLabel(code);
      
      if (lbl.length() == 0 || !lbl.endsWith("-" + lblOrig))
      {
         // return original code if there is no label or if the label doesn't end with the original
         // key
         return orig;
      }
      
      return code;
   }

   /**
    * Returns <code>true</code> if, for the given key, a {@link #keyTyped} event is also raised,
    * beside a {@link #keyPressed}, when used in conjunction with the Control key.
    * 
    * @param    key
    *           The pressed key.
    *           
    * @return   <code>true</code> if the <code>CTRL-<key></code> combination will be  treated by
    *           {@link #keyTyped(KeyEvent)}.
    */
   private boolean ctrlTyped(int key)
   {
      return (key >= 'a' && key <= 'z') || (key >= 'A' && key <= 'Z');
   }
}
